@@ -1,7 +1,8 @@
 from .tsrestapiv1 import TSRestApiV1
 from .tsrestapiv1 import (
     TSTypes, Sorts, Categories,
-    ShareModes, Privileges, PermissionTypes, MetadataTypes, MetadataSubtypes
+    ShareModes, Privileges, PermissionTypes, MetadataTypes, MetadataSubtypes, GroupVisibility
 )
 from .tsrestapiv2 import TSRestApiV2, ReportTypes, TSTypesV2
+from .details_objects import *
 from ._version import __version__
