@@ -127,6 +127,12 @@ class PermissionTypes:
     DEFINED = 'DEFINED'
 
 
+class GroupVisibility:
+    DEFAULT = 'DEFAULT'
+    SHARABLE = 'DEFAULT'
+    NON_SHARABLE = 'NON_SHARABLE'
+
+
 # Compatibility
 MetadataNames = MetadataTypes
 MetadataSorts = Sorts
@@ -170,13 +176,16 @@ class TSRestApiV1:
         # V2 REST API for basic implementation
         self.v2_base_url = '{server}/tspublic/rest/v2/'.format(server=self.server)
 
+        # Flag for whether the version implements the export_fqn option of metadata/tml/export
+        self.can_export_fqn = True
+
     #
     # Session management calls
     # - up here vs. in the SESSION section below (because these two are required)
     #
-    def session_login(self, username: str, password: str) -> bool:
+    def session_login(self, username: str, password: str, remember_me: bool = True) -> bool:
         endpoint = 'session/login'
-        post_data = {'username': username, 'password': password, 'rememberme': 'true'}
+        post_data = {'username': username, 'password': password, 'rememberme': str(remember_me).lower()}
 
         url = self.base_url + endpoint
         response = self.requests_session.post(url=url, data=post_data)
@@ -195,55 +204,6 @@ class TSRestApiV1:
         response.raise_for_status()
         return True
 
-    #
-    # V2 Session management
-    #
-
-    def session_login_v2(self,  username: Optional[str] = None, password: Optional[str] = None,
-                         token: Optional[str] = None) -> bool:
-        endpoint = 'session/login'
-
-        url = self.v2_base_url + endpoint
-
-        if token is not None:
-            response = self.requests_session.post(url=url, headers={"Authorization": "Bearer {}".format(token)})
-        elif username is not None and password is not None:
-            json_post_data = {'userName': username, 'password': password, 'rememberMe': 'true'}
-            response = self.requests_session.post(url=url, json=json_post_data)
-        else:
-            raise Exception("If using username/password, must include both")
-
-        # HTTP 204 - success, no content
-        response.raise_for_status()
-        return True
-
-    # This is only the V2 API bearer token, not the Trusted Authentication token from V1
-    def get_token_v2(self,  username: Optional[str] = None, password: Optional[str] = None,
-                     secret_key: Optional[str] = None, token_expiry_duration: int = 300,
-                     access_level: str = "FULL", ts_object_id: Optional[str] = None) -> Dict:
-        endpoint = 'session/gettoken'
-
-        url = self.v2_base_url + endpoint
-
-        json_post_data = {'accessLevel': access_level,
-                          'tokenExpiryDuration': token_expiry_duration}
-
-        if secret_key is not None:
-            json_post_data['secretKey'] = secret_key
-        elif username is not None and password is not None:
-            json_post_data['userName'] = username
-            json_post_data['password'] = password
-        else:
-            raise Exception("If using username/password, must include both")
-
-        if ts_object_id is not None:
-            json_post_data['tsObjectId'] = ts_object_id
-
-        response = self.requests_session.post(url=url, json=json_post_data, headers={'Accept-Language': 'en_US'})
-
-        response.raise_for_status()
-        return response.json()
-
     #
     # Root level API methods found below, divided into logical separations
     #
@@ -886,6 +846,16 @@ class TSRestApiV1:
         response.raise_for_status()
         return response.json()
 
+    # Helper method to find a GUID from a name
+    def metadata_list_find_guid(self, object_type: str, name: str):
+        objects = self.metadata_list(object_type=object_type, filter=name)
+        # Filter is case-insensitive and the equivalent of a wild-card, so need to look for exact match
+        # on the response
+        for o in objects['headers']:
+            if o['name'] == name:
+                return o['id']
+        raise LookupError()
+
     # Favorite Methods
     def metadata_markasfavoritefor(self, user_guid: str, object_guids: List[str], object_type: str) -> Dict:
         endpoint = 'metadata/markunmarkfavoritefor'
@@ -947,7 +917,7 @@ class TSRestApiV1:
             else:
                 return response.json()
 
-    def metadata_tml_export(self, guid: str, export_associated=False) -> OrderedDict:
+    def metadata_tml_export(self, guid: str, export_associated=False, export_fqn=True) -> OrderedDict:
         # Always returns a Python Dict, converted from a request to the API to receive in JSON
         endpoint = 'metadata/tml/export'
 
@@ -956,6 +926,9 @@ class TSRestApiV1:
             'formattype': 'JSON',
             'export_associated': str(export_associated).lower()
         }
+        # Added in version 8.9, can be set to skip for older releases
+        if self.can_export_fqn is True:
+            post_data['export_fqn'] = str(export_fqn).lower()
 
         url = self.base_url + endpoint
         # TML import is distinguished by having an {'Accept': 'text/plain'} header on the POST
@@ -981,7 +954,7 @@ class TSRestApiV1:
         return tml_obj
 
     # Method to retrieve the all of the associate objects and retrieve original object and a dict of name:guid mapping
-    def metadata_tml_export_with_associations_map(self, guid: str) -> (OrderedDict, Dict):
+    def metadata_tml_export_with_associations_map(self, guid: str, export_fqn=True) -> (OrderedDict, Dict):
         # Always returns a Python Dict, converted from a request to the API to receive in JSON
         endpoint = 'metadata/tml/export'
 
@@ -990,6 +963,9 @@ class TSRestApiV1:
             'formattype': 'JSON',
             'export_associated': 'true'
         }
+        # Added in version 8.9, can be set to skip for older releases
+        if self.can_export_fqn is True:
+            post_data['export_fqn'] = str(export_fqn).lower()
 
         url = self.base_url + endpoint
         # TML import is distinguished by having an {'Accept': 'text/plain'} header on the POST
@@ -1014,7 +990,8 @@ class TSRestApiV1:
 
         return tml_obj, name_guid_map
 
-    def metadata_tml_export_string(self, guid: str, formattype: str = 'YAML', export_associated=False) -> str:
+    def metadata_tml_export_string(self, guid: str, formattype: str = 'YAML',
+                                   export_associated=False, export_fqn=True) -> str:
         # Intended for a direct pull with no conversion
         endpoint = 'metadata/tml/export'
         # allow JSON or YAML in any casing
@@ -1024,6 +1001,9 @@ class TSRestApiV1:
             'formattype': formattype,
             'export_associated': str(export_associated).lower()
         }
+        # Added in version 8.9, can be set to skip for older releases
+        if self.can_export_fqn is True:
+            post_data['export_fqn'] = str(export_fqn).lower()
         url = self.base_url + endpoint
 
         # TML import is distinguished by having an {'Accept': 'text/plain'} header on the POST
@@ -1046,7 +1026,8 @@ class TSRestApiV1:
         else:
             raise Exception()
 
-    def metadata_tml_export_string_with_associations_map(self, guid: str, formattype: str = 'YAML') -> (str, Dict):
+    def metadata_tml_export_string_with_associations_map(self, guid: str, formattype: str = 'YAML',
+                                                         export_fqn=True) -> (str, Dict):
         # Intended for a direct pull with no conversion
         endpoint = 'metadata/tml/export'
         # allow JSON or YAML in any casing
@@ -1056,6 +1037,9 @@ class TSRestApiV1:
             'formattype': formattype,
             'export_associated': 'true'
         }
+        # Added in version 8.9, can be set to skip for older releases
+        if self.can_export_fqn is True:
+            post_data['export_fqn'] = str(export_fqn).lower()
         url = self.base_url + endpoint
 
         # TML import is distinguished by having an {'Accept': 'text/plain'} header on the POST
@@ -1084,7 +1068,8 @@ class TSRestApiV1:
         tml: Union[Dict, List[Dict]],
         create_new_on_server: bool = False,
         validate_only: bool = False,
-        formattype: str = 'JSON'
+        formattype: str = 'JSON',
+        enable_block_tml_metadata_sync: Optional[bool] = None
     ) -> Dict:
         endpoint = 'metadata/tml/import'
         # allow JSON or YAML in any casing
@@ -1114,6 +1099,8 @@ class TSRestApiV1:
             'import_policy': import_policy,
             'force_create': str(create_new_on_server).lower()
         }
+        if enable_block_tml_metadata_sync is not None:
+            post_data['enable_block_tml_metadata_sync'] = str(enable_block_tml_metadata_sync).lower()
 
         url = self.base_url + endpoint
 
@@ -1166,6 +1153,19 @@ class TSRestApiV1:
             permissions_dict[l1][guid] = {'shareMode': share_mode}
         return permissions_dict
 
+    # Helper to create share permissions
+    def create_share_permissions(self, read_only_users_or_groups_guids: Optional[List[str]] = (),
+                                 edit_access_users_or_groups_guids: Optional[List[str]] = (),
+                                 remove_access_users_or_groups_groups: Optional[List[str]] = ()) -> Dict:
+        permissions_dict = self.get_sharing_permissions_dict()
+        for a in read_only_users_or_groups_guids:
+            self.add_permission_to_dict(permissions_dict=permissions_dict, guid=a, share_mode=ShareModes.READ_ONLY)
+        for a in edit_access_users_or_groups_guids:
+            self.add_permission_to_dict(permissions_dict=permissions_dict, guid=a, share_mode=ShareModes.EDIT)
+        for a in remove_access_users_or_groups_groups:
+            self.add_permission_to_dict(permissions_dict=permissions_dict, guid=a, share_mode=ShareModes.NO_ACCESS)
+        return permissions_dict
+
     # Share any object type
     # Requires a Permissions Dict, which can be generated and modified with the two static methods above
     def security_share(
@@ -1182,7 +1182,9 @@ class TSRestApiV1:
             email_shares = []
 
         endpoint = 'security/share'
-
+        if shared_object_type in [MetadataSubtypes.TABLE, MetadataSubtypes.VIEW, MetadataSubtypes.WORKSHEET,
+                                  MetadataSubtypes.SQL_VIEW]:
+            shared_object_type = MetadataTypes.TABLE
         post_data = {
             'type': shared_object_type,
             'id': json.dumps(shared_object_guids),
@@ -1311,29 +1313,99 @@ class TSRestApiV1:
         response.raise_for_status()
         return True
 
+    def session_info(self) -> Dict:
+        endpoint = 'session/info'
+        url = self.base_url + endpoint
+        response = self.requests_session.get(url=url)
+        response.raise_for_status()
+        return response.json()
+
+    def session_orgs_get(self, batchsize: int = -1, offset: int = -1) -> Dict:
+        endpoint = 'session/orgs'
+
+        url_params = {
+            'batchsize': batchsize,
+            'offset': offset
+        }
+
+        url = self.base_url + endpoint
+        response = self.requests_session.get(url=url, params=url_params)
+        response.raise_for_status()
+        return response.json()
+
+    def session_orgs_put(self, org_id: int):
+        endpoint = 'session/orgs'
+
+        post_data = {
+            'orgid': org_id
+        }
+
+        url = self.base_url + endpoint
+        response = self.requests_session.put(url=url, data=post_data)
+        response.raise_for_status()
+
+    def session_orgs_users(self, user_guid: str, org_scope: str = 'ALL'):
+        endpoint = 'session/orgs/users/{}'.format(user_guid)
+
+        url_params = {
+            'orgScope': org_scope
+        }
+
+        url = self.base_url + endpoint
+        response = self.requests_session.get(url=url, params=url_params)
+        response.raise_for_status()
+        return response.json()
+
     # NOTE:
     #
-    #   session/login/token is not implemented here, it is intended for a browser login
-    #
     #   The below shows an implementation of session/auth/token but it should only be
     #   used from Authenticator Server with Secret Key retrieved in a secure manner only
     #   in memory
     #
-    def session_auth_token(self, secret_key: str, username: str, access_level: str, object_id: str):
+    def session_auth_token(self, secret_key: str, username: str, access_level: str = 'FULL',
+                           object_guid: Optional[str] = None, org_id: Optional[int] = None,
+                           groups: Optional[List[str]] = None, auto_create_user: Optional[bool] = None):
         endpoint = 'session/auth/token'
 
         post_params = {
              'secret_key': secret_key,
              'username': username,
              'access_level': access_level,
-             'id': object_id
+
         }
+        if object_guid is not None:
+            post_params['id'] = object_guid
+
+        if org_id is not None:
+            post_params['orgid'] = org_id
+
+        if auto_create_user is not None:
+            post_params['autocreate'] = str(auto_create_user).lower()
+
+        if groups is not None:
+            post_params['groups'] = json.dumps(groups)
 
         url = self.base_url + endpoint
 
         response = self.requests_session.post(url=url, data=post_params)
         return response
 
+    # session/login/token is typically only used within the browser and handled by the Visual Embed SDK,
+    # provided here for testing
+    def session_login_token_post(self, username: str, auth_token: str, redirect_url: str):
+        endpoint = 'session/login/token'
+
+        post_params = {
+             'username': username,
+             'auth_token': auth_token,
+             'redirect_url': redirect_url,  # need to url encode
+
+        }
+
+        url = self.base_url + endpoint
+
+        response = self.requests_session.post(url=url, data=post_params)
+        return response
     #
     # USER Methods
     #
@@ -1351,7 +1423,7 @@ class TSRestApiV1:
         response.raise_for_status()
         return response.json()
 
-    def user_post(self, username: str, password: str, display_name: str, properties: Optional,
+    def user_post(self, username: str, password: str, display_name: str, properties: Optional[Dict] = None,
                   groups: Optional[List[str]] = None, user_type: str = 'LOCAL_USER',
                   tenant_id: Optional[str] = None, visibility: str = 'DEFAULT'):
         endpoint = 'user'
@@ -1452,7 +1524,7 @@ class TSRestApiV1:
         return response.json()
 
     def user_transfer_ownership(self, current_owner_username: str, new_owner_username: str,
-                                object_guids: Optional[List[str]] = None) -> Dict:
+                                object_guids: Optional[List[str]] = None) -> bool:
         endpoint = 'user/transfer/ownership'
 
         url_params = {
@@ -1466,7 +1538,7 @@ class TSRestApiV1:
         url = self.base_url + endpoint
         response = self.requests_session.post(url=url, params=url_params)
         response.raise_for_status()
-        return response.json()
+        return True
 
     # NOTE: preferences and preferencesProto are a big ?
     def user_updatepreference(self, user_guid: str, username: str, preferences: Dict,
